import tempfile
import webbrowser
from collections.abc import Iterator, Mapping, Sequence, Set
from typing import Any, Union

import markdown
import yaml

from dagster_dg.component import RemoteComponentType
from dagster_dg.yaml_utils import parse_yaml_with_source_positions
from dagster_dg.yaml_utils.source_position import SourcePositionTree

REF_BASE = "#/$defs/"
JSON_SCHEMA_EXTRA_REQUIRED_SCOPE_KEY = "dagster_required_scope"


def _subschemas_on_path(
    valpath: Sequence[Union[str, int]], json_schema: Mapping[str, Any], subschema: Mapping[str, Any]
) -> Iterator[Mapping[str, Any]]:
    """Given a valpath and the json schema of a given target type, returns the subschemas at each step of the path."""
    # List[ComplexType] (e.g.) will contain a reference to the complex type schema in the
    # top-level $defs, so we dereference it here.
    if "$ref" in subschema:
        # depending on the pydantic version, the extras may be stored with the reference or not
        extras = {k: v for k, v in subschema.items() if k != "$ref"}
        subschema = {**json_schema["$defs"].get(subschema["$ref"][len(REF_BASE) :]), **extras}

    yield subschema
    if len(valpath) == 0:
        return

    # Optional[ComplexType] (e.g.) will contain multiple schemas in the "anyOf" field
    if "anyOf" in subschema:
        for inner in subschema["anyOf"]:
            yield from _subschemas_on_path(valpath, json_schema, inner)

    el = valpath[0]
    if isinstance(el, str):
        # valpath: ['field']
        # field: X
        inner = subschema.get("properties", {}).get(el)
    elif isinstance(el, int):
        # valpath: ['field', 0]
        # field: List[X]
        inner = subschema.get("items")
    else:
        raise ValueError(f"Invalid valpath element: {el}")

    # the path wasn't valid, or unspecified
    if not inner:
        return

    _, *rest = valpath
    yield from _subschemas_on_path(rest, json_schema, inner)


def _get_additional_required_scope(subschema: Mapping[str, Any]) -> Set[str]:
    raw = subschema.get(JSON_SCHEMA_EXTRA_REQUIRED_SCOPE_KEY)
    return set(raw) if raw else set()


def get_required_scope(
    valpath: Sequence[Union[str, int]], json_schema: Mapping[str, Any]
) -> Set[str]:
    """Given a valpath and the json schema of a given target type, determines the available rendering scope."""
    required_scope = set()
    for subschema in _subschemas_on_path(valpath, json_schema, json_schema):
        required_scope |= _get_additional_required_scope(subschema)
    return required_scope


def _dereference_schema(
    json_schema: Mapping[str, Any], subschema: Mapping[str, Any]
) -> Mapping[str, Any]:
    if "$ref" in subschema:
        return json_schema["$defs"].get(subschema["$ref"][len(REF_BASE) :])
    else:
        return subschema


def _sample_value_for_subschema(
    json_schema: Mapping[str, Any], subschema: Mapping[str, Any]
) -> Any:
    subschema = _dereference_schema(json_schema, subschema)
    if "anyOf" in subschema:
        # TODO: handle anyOf fields more gracefully, for now just choose first option
        return _sample_value_for_subschema(json_schema, subschema["anyOf"][0])

    objtype = subschema["type"]
    if objtype == "object":
        return {
            k: _sample_value_for_subschema(json_schema, v)
            for k, v in subschema.get("properties", {}).items()
        }
    elif objtype == "array":
        return [_sample_value_for_subschema(json_schema, subschema["items"])]
    elif objtype == "string":
        return "..."
    elif objtype == "integer":
        return 0
    elif objtype == "boolean":
        return False
    else:
        return f"({objtype})"


class ComponentDumper(yaml.SafeDumper):
    def increase_indent(self, flow=False, *args, **kwargs):
        # makes the output somewhat prettier by forcing lists to be indented
        return super().increase_indent(flow=flow, indentless=False)

    def write_line_break(self) -> None:
        # add an extra line break between top-level keys
        if self.indent == 0:
            super().write_line_break()
        super().write_line_break()


def _get_source_position_comments(
    valpath: Sequence[Union[str, int]], tree: SourcePositionTree, json_schema: Mapping[str, Any]
) -> Iterator[tuple[int, str]]:
    available_scope = get_required_scope(valpath[1:], json_schema)
    if available_scope:
        yield (tree.position.start.line - 1, f"Available scope: {available_scope}")
    for child_path, child_tree in tree.children.items():
        yield from _get_source_position_comments([*valpath, child_path], child_tree, json_schema)


def generate_sample_yaml(component_type: str, json_schema: Mapping[str, Any]) -> str:
    raw = yaml.dump(
        {"type": component_type, "params": _sample_value_for_subschema(json_schema, json_schema)},
        Dumper=ComponentDumper,
        sort_keys=False,
    )
    parsed = parse_yaml_with_source_positions(raw)
    comments = dict(_get_source_position_comments([], parsed.source_position_tree, json_schema))
    commented_lines = []
    for line_num, line in enumerate(raw.split("\n")):
        if line_num in comments:
            commented_lines.append(f"{line} # {comments[line_num]}")
        else:
            commented_lines.append(line)
    return "\n".join(commented_lines)


def render_markdown_in_browser(markdown_content: str) -> None:
    # Convert the markdown string to HTML
    html_content = markdown.markdown(markdown_content)

    # Add basic HTML structure
    full_html = f"""
    <!DOCTYPE html>
    <html lang="en">
    <head>
        <meta charset="UTF-8">
        <meta name="viewport" content="width=device-width, initial-scale=1.0">
        <title>Markdown Preview</title>
    </head>
    <body>
        {html_content}
    </body>
    </html>
    """

    # Create a temporary file
    with tempfile.NamedTemporaryFile(delete=False, suffix=".html") as temp_file:
        temp_file.write(full_html.encode("utf-8"))
        temp_file_path = temp_file.name

    # Open the temporary file in the default web browser
    webbrowser.open(f"file://{temp_file_path}")


def markdown_for_component_type(remote_component_type: RemoteComponentType) -> str:
    component_type_name = f"{remote_component_type.package}.{remote_component_type.name}"
    sample_yaml = generate_sample_yaml(
        component_type_name, remote_component_type.component_params_schema or {}
    )
    rows = len(sample_yaml.split("\n")) + 1
    return f"""
## Component: `{component_type_name}`
 
### Description: 
{remote_component_type.description}

### Sample Component Params:

<textarea rows={rows} cols=100>
{sample_yaml}
</textarea>
"""



import os
from pathlib import Path

import click

from dagster_dg.cli.global_options import dg_global_options
from dagster_dg.config import normalize_cli_config
from dagster_dg.context import DgContext
from dagster_dg.scaffold import scaffold_deployment
from dagster_dg.utils import DgClickCommand, DgClickGroup, exit_with_error


@click.group(name="deployment", cls=DgClickGroup)
def deployment_group():
    """Commands for operating on deployment directories."""


# ########################
# ##### SCAFFOLD
# ########################


@deployment_group.command(name="scaffold", cls=DgClickCommand)
@dg_global_options
@click.argument("path", type=Path)
@click.pass_context
def deployment_scaffold_command(
    context: click.Context, path: Path, **global_options: object
) -> None:
    """Scaffold a Dagster deployment file structure.

    The deployment file structure includes a directory for code locations and configuration files
    for deploying to Dagster Plus.
    """
    cli_config = normalize_cli_config(global_options, context)
    dg_context = DgContext.from_config_file_discovery_and_cli_config(Path.cwd(), cli_config)
    dir_abspath = os.path.abspath(path)
    if os.path.exists(dir_abspath):
        exit_with_error(
            f"A file or directory at {dir_abspath} already exists. "
            + "\nPlease delete the contents of this path or choose another location."
        )
    scaffold_deployment(path, dg_context)



import copy
import json
from collections import defaultdict
from collections.abc import Iterable, Mapping, Sequence
from dataclasses import dataclass
from pathlib import Path
from typing import TYPE_CHECKING, Any, Optional

from dagster_dg.utils import is_valid_json

if TYPE_CHECKING:
    from dagster_dg.context import DgContext


@dataclass
class RemoteComponentType:
    name: str
    package: str
    summary: Optional[str]
    description: Optional[str]
    scaffold_params_schema: Optional[Mapping[str, Any]]  # json schema
    component_params_schema: Optional[Mapping[str, Any]]  # json schema

    @property
    def key(self) -> str:
        return self.name


def all_components_schema_from_dg_context(dg_context: "DgContext") -> Mapping[str, Any]:
    """Generate a schema for all components in the current environment, or retrieve it from the cache."""
    schema_raw = None
    if dg_context.has_cache:
        cache_key = dg_context.get_cache_key("all_components_schema")
        schema_raw = dg_context.cache.get(cache_key)

    if not schema_raw:
        schema_raw = dg_context.external_components_command(["list", "all-components-schema"])
    return json.loads(schema_raw)


def _retrieve_local_component_types(
    dg_context: "DgContext", paths: Sequence[Path]
) -> Mapping[str, Mapping[str, Mapping[str, Any]]]:
    paths_to_fetch = set(paths)
    data_for_path: dict[str, dict[str, Mapping[str, Any]]] = {}
    if dg_context.has_cache:
        for path in paths:
            cache_key = dg_context.get_cache_key_for_local_components(path)
            raw_data = dg_context.cache.get(cache_key)
            if raw_data:
                data_for_path[str(path)] = json.loads(raw_data)
                paths_to_fetch.remove(path)

    if paths_to_fetch:
        raw_local_component_data = dg_context.external_components_command(
            [
                "list",
                "local-component-types",
                *[str(path) for path in paths_to_fetch],
            ]
        )
        local_component_data = json.loads(raw_local_component_data)
        for path in paths_to_fetch:
            data_for_path[str(path)] = local_component_data.get(str(path), {})

        if dg_context.has_cache:
            for path in paths_to_fetch:
                cache_key = dg_context.get_cache_key_for_local_components(path)
                data_for_path_json = json.dumps(local_component_data.get(str(path), {}))
                if cache_key and is_valid_json(data_for_path_json):
                    dg_context.cache.set(cache_key, data_for_path_json)

    return data_for_path


class RemoteComponentRegistry:
    @classmethod
    def from_dg_context(
        cls, dg_context: "DgContext", local_component_type_dirs: Optional[Sequence[Path]] = None
    ) -> "RemoteComponentRegistry":
        """Fetches the set of available component types, including local component types for the
        specified directories. Caches the result if possible.
        """
        if dg_context.use_dg_managed_environment:
            dg_context.ensure_uv_lock()

        if dg_context.has_cache:
            cache_key = dg_context.get_cache_key("component_registry_data")
            raw_registry_data = dg_context.cache.get(cache_key)
        else:
            cache_key = None
            raw_registry_data = None

        if not raw_registry_data:
            raw_registry_data = dg_context.external_components_command(["list", "component-types"])
            if dg_context.has_cache and cache_key and is_valid_json(raw_registry_data):
                dg_context.cache.set(cache_key, raw_registry_data)

        registry_data = json.loads(raw_registry_data)

        local_component_data: Mapping[str, Mapping[str, Mapping[str, Any]]] = {}
        if local_component_type_dirs:
            local_component_data = _retrieve_local_component_types(
                dg_context, local_component_type_dirs
            )
        return cls.from_dict(
            global_component_types=registry_data, local_component_types=local_component_data
        )

    @classmethod
    def from_dict(
        cls,
        global_component_types: Mapping[str, Mapping[str, Any]],
        local_component_types: Mapping[str, Mapping[str, Mapping[str, Any]]],
    ) -> "RemoteComponentRegistry":
        components_by_path = defaultdict(dict)
        for directory in local_component_types:
            for key, metadata in local_component_types[directory].items():
                components_by_path[directory][key] = RemoteComponentType(**metadata)

        return RemoteComponentRegistry(
            {key: RemoteComponentType(**value) for key, value in global_component_types.items()},
            local_components=components_by_path,
        )

    def __init__(
        self,
        components: dict[str, RemoteComponentType],
        local_components: dict[str, dict[str, RemoteComponentType]],
    ):
        self._components: dict[str, RemoteComponentType] = copy.copy(components)
        self._local_components: dict[str, dict[str, RemoteComponentType]] = copy.copy(
            local_components
        )

    @staticmethod
    def empty() -> "RemoteComponentRegistry":
        return RemoteComponentRegistry({}, {})

    def has_global(self, name: str) -> bool:
        return name in self._components

    def get(self, path: Path, key: str) -> RemoteComponentType:
        """Resolves a component type within the scope of a given component directory."""
        if key in self._components:
            return self._components[key]

        return self._local_components[str(path)][key]

    def get_global(self, name: str) -> RemoteComponentType:
        return self._components[name]

    def global_keys(self) -> Iterable[str]:
        return self._components.keys()

    def global_items(self) -> Iterable[tuple[str, RemoteComponentType]]:
        for key in sorted(self.global_keys()):
            yield key, self.get_global(key)

    def __repr__(self) -> str:
        return f"<RemoteComponentRegistry {list(self._components.keys())}>"



from collections.abc import Sequence
from pathlib import Path
from typing import Any, Callable, Optional, TypeVar, Union

import click

from dagster_dg.config import DgConfig
from dagster_dg.utils import set_option_help_output_group

T_Command = TypeVar("T_Command", bound=Union[Callable[..., Any], click.Command])

# Defaults are defined on the DgConfig object.
GLOBAL_OPTIONS = {
    option.name: option
    for option in [
        click.Option(
            ["--cache-dir"],
            type=Path,
            default=DgConfig.cache_dir,
            help="Specify a directory to use for the cache.",
        ),
        click.Option(
            ["--disable-cache"],
            is_flag=True,
            default=DgConfig.disable_cache,
            help="Disable the cache..",
        ),
        click.Option(
            ["--verbose"],
            is_flag=True,
            default=DgConfig.verbose,
            help="Enable verbose output for debugging.",
        ),
        click.Option(
            ["--builtin-component-lib"],
            type=str,
            default=DgConfig.builtin_component_lib,
            help="Specify a builitin component library to use.",
        ),
        click.Option(
            ["--use-dg-managed-environment/--no-use-dg-managed-environment"],
            is_flag=True,
            default=DgConfig.use_dg_managed_environment,
            help="Enable management of the virtual environment with uv.",
        ),
    ]
}

# Ensure that these options show up in the help output under the "Global options" group.
for option in GLOBAL_OPTIONS.values():
    set_option_help_output_group(option, "Global options")


def dg_global_options(
    fn: Optional[T_Command] = None, *, names: Optional[Sequence[str]] = None
) -> Union[T_Command, Callable[[T_Command], T_Command]]:
    if fn:
        options = [GLOBAL_OPTIONS[name] for name in names or list(GLOBAL_OPTIONS.keys())]
        if isinstance(fn, click.Command):
            for option in options:
                fn.params.append(option)
        else:
            # This is borrowed from click itself, it is how its decorators operate on both commands
            # and regular functions.
            if not hasattr(fn, "__click_params__"):
                fn.__click_params__ = []  # type: ignore
            for option in options:
                fn.__click_params__.append(option)  # type: ignore

        return fn
    else:
        return lambda fn: dg_global_options(fn, names=names)  # type: ignore


def validate_global_opts(context: click.Context, **global_options: object) -> None:
    for name, value in global_options.items():
        if name not in GLOBAL_OPTIONS:
            raise click.UsageError(f"Unknown global option: {name}")
        GLOBAL_OPTIONS[name].process_value(context, value)



from collections.abc import Mapping, Sequence
from copy import copy
from pathlib import Path
from typing import Any, NamedTuple, Optional

import click
from click.core import ParameterSource
from jsonschema import Draft202012Validator, ValidationError
from typer.rich_utils import rich_format_help
from yaml.scanner import ScannerError

from dagster_dg.cli.check_utils import error_dict_to_formatted_error
from dagster_dg.cli.global_options import GLOBAL_OPTIONS, dg_global_options
from dagster_dg.component import RemoteComponentRegistry, RemoteComponentType
from dagster_dg.config import (
    get_config_from_cli_context,
    has_config_on_cli_context,
    normalize_cli_config,
    set_config_on_cli_context,
)
from dagster_dg.context import DgContext
from dagster_dg.scaffold import scaffold_component_instance
from dagster_dg.utils import (
    DgClickCommand,
    DgClickGroup,
    exit_with_error,
    json_schema_property_to_click_option,
    not_none,
    parse_json_option,
)
from dagster_dg.yaml_utils import parse_yaml_with_source_positions
from dagster_dg.yaml_utils.source_position import (
    LineCol,
    SourcePosition,
    SourcePositionTree,
    ValueAndSourcePositionTree,
)


@click.group(name="component", cls=DgClickGroup)
def component_group():
    """Commands for operating on components."""


# ########################
# ##### SCAFFOLD
# ########################


# The `dg component scaffold` command is special because its subcommands are dynamically generated
# from the registered component types in the code location. Because the registered component types
# depend on the built-in component library we are using, we cannot resolve them until we have the
# built-in component library, which can be set via a global option, e.g.:
#
#     dg --builtin-component-lib dagster_components.test ...
#
# To handle this, we define a custom click.Group subclass that loads the commands on demand.
class ComponentScaffoldGroup(DgClickGroup):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._commands_defined = False

    def get_command(self, cli_context: click.Context, cmd_name: str) -> Optional[click.Command]:
        if not self._commands_defined:
            self._define_commands(cli_context)
        return super().get_command(cli_context, cmd_name)

    def list_commands(self, cli_context: click.Context) -> list[str]:
        if not self._commands_defined:
            self._define_commands(cli_context)
        return super().list_commands(cli_context)

    def _define_commands(self, cli_context: click.Context) -> None:
        """Dynamically define a command for each registered component type."""
        if not has_config_on_cli_context(cli_context):
            cli_context.invoke(not_none(self.callback), **cli_context.params)
        config = get_config_from_cli_context(cli_context)
        dg_context = DgContext.from_config_file_discovery_and_cli_config(Path.cwd(), config)

        if not dg_context.is_code_location:
            exit_with_error("This command must be run inside a Dagster code location directory.")

        registry = RemoteComponentRegistry.from_dg_context(dg_context)
        for key, component_type in registry.global_items():
            command = _create_component_scaffold_subcommand(key, component_type)
            self.add_command(command)


class ComponentScaffoldSubCommand(DgClickCommand):
    # We have to override this because the implementation of `format_help` used elsewhere will only
    # pull parameters directly off the target command. For these component scaffold subcommands  we need
    # to expose the global options, which are defined on the preceding group rather than the command
    # itself.
    def format_help(self, context: click.Context, formatter: click.HelpFormatter):
        """Customizes the help to include hierarchical usage."""
        if not isinstance(self, click.Command):
            raise ValueError("This mixin is only intended for use with click.Command instances.")

        # This is a hack. We pass the help format func a modified version of the command where the global
        # options are attached to the command itself. This will cause them to be included in the
        # help output.
        cmd_copy = copy(self)
        cmd_copy.params = [
            *cmd_copy.params,
            *(GLOBAL_OPTIONS.values()),
        ]
        rich_format_help(obj=cmd_copy, ctx=context, markup_mode="rich")

    def format_usage(self, context: click.Context, formatter: click.HelpFormatter) -> None:
        if not isinstance(self, click.Command):
            raise ValueError("This mixin is only intended for use with click.Command instances.")
        arg_pieces = self.collect_usage_pieces(context)
        command_parts = context.command_path.split(" ")
        command_parts.insert(-1, "[GLOBAL OPTIONS]")
        return formatter.write_usage(" ".join(command_parts), " ".join(arg_pieces))


# We have to override the usual Click processing of `--help` here. The issue is
# that click will process this option before processing anything else, but because we are
# dynamically generating subcommands based on the content of other options, the output of --help
# actually depends on these other options. So we opt out of Click's short-circuiting
# behavior of `--help` by setting `help_option_names=[]`, ensuring that we can process the other
# options first and generate the correct subcommands. We then add a custom `--help` option that
# gets invoked inside the callback.
@component_group.group(
    name="scaffold",
    cls=ComponentScaffoldGroup,
    invoke_without_command=True,
    context_settings={"help_option_names": []},
)
@click.option("-h", "--help", "help_", is_flag=True, help="Show this message and exit.")
@dg_global_options
@click.pass_context
def component_scaffold_group(context: click.Context, help_: bool, **global_options: object) -> None:
    """Scaffold of a Dagster component."""
    # Click attempts to resolve subcommands BEFORE it invokes this callback.
    # Therefore we need to manually invoke this callback during subcommand generation to make sure
    # it runs first. It will be invoked again later by Click. We make it idempotent to deal with
    # that.
    if not has_config_on_cli_context(context):
        cli_config = normalize_cli_config(global_options, context)
        set_config_on_cli_context(context, cli_config)
    if help_:
        click.echo(context.get_help())
        context.exit(0)


def _create_component_scaffold_subcommand(
    component_key: str, component_type: RemoteComponentType
) -> DgClickCommand:
    # We need to "reset" the help option names to the default ones because we inherit the parent
    # value of context settings from the parent group, which has been customized.
    @click.command(
        name=component_key,
        cls=ComponentScaffoldSubCommand,
        context_settings={"help_option_names": ["-h", "--help"]},
    )
    @click.argument("component_name", type=str)
    @click.option(
        "--json-params",
        type=str,
        default=None,
        help="JSON string of component parameters.",
        callback=parse_json_option,
    )
    @click.pass_context
    def scaffold_component_command(
        cli_context: click.Context,
        component_name: str,
        json_params: Mapping[str, Any],
        **key_value_params: Any,
    ) -> None:
        f"""Scaffold of a {component_type.name} component.

        This command must be run inside a Dagster code location directory. The component scaffold will be
        placed in submodule `<code_location_name>.components.<COMPONENT_NAME>`.

        Components can optionally be passed scaffold parameters. There are two ways to do this:

        (1) Passing a single --json-params option with a JSON string of parameters. For example:

            dg component scaffold foo.bar my_component --json-params '{{"param1": "value", "param2": "value"}}'`.

        (2) Passing each parameter as an option. For example:

            dg component scaffold foo.bar my_component --param1 value1 --param2 value2`

        It is an error to pass both --json-params and key-value pairs as options.
        """
        cli_config = get_config_from_cli_context(cli_context)
        dg_context = DgContext.for_code_location_environment(Path.cwd(), cli_config)

        registry = RemoteComponentRegistry.from_dg_context(dg_context)
        if not registry.has_global(component_key):
            exit_with_error(f"No component type `{component_key}` could be resolved.")
        elif dg_context.has_component(component_name):
            exit_with_error(f"A component instance named `{component_name}` already exists.")

        # Specified key-value params will be passed to this function with their default value of
        # `None` even if the user did not set them. Filter down to just the ones that were set by
        # the user.
        user_provided_key_value_params = {
            k: v
            for k, v in key_value_params.items()
            if cli_context.get_parameter_source(k) == ParameterSource.COMMANDLINE
        }
        if json_params is not None and user_provided_key_value_params:
            exit_with_error(
                "Detected params passed as both --json-params and individual options. These are mutually exclusive means of passing"
                " component generation parameters. Use only one.",
            )
        elif json_params:
            scaffold_params = json_params
        elif user_provided_key_value_params:
            scaffold_params = user_provided_key_value_params
        else:
            scaffold_params = None

        scaffold_component_instance(
            Path(dg_context.components_path) / component_name,
            component_key,
            scaffold_params,
            dg_context,
        )

    # If there are defined scaffold params, add them to the command
    schema = component_type.scaffold_params_schema
    if schema:
        for key, field_info in schema["properties"].items():
            # All fields are currently optional because they can also be passed under
            # `--json-params`
            option = json_schema_property_to_click_option(key, field_info, required=False)
            scaffold_component_command.params.append(option)

    return scaffold_component_command


# ########################
# ##### LIST
# ########################


@component_group.command(name="list", cls=DgClickCommand)
@dg_global_options
@click.pass_context
def component_list_command(context: click.Context, **global_options: object) -> None:
    """List Dagster component instances defined in the current code location."""
    cli_config = normalize_cli_config(global_options, context)
    dg_context = DgContext.for_code_location_environment(Path.cwd(), cli_config)

    for component_name in dg_context.get_component_names():
        click.echo(component_name)


# ########################
# ##### CHECK
# ########################

COMPONENT_FILE_SCHEMA = {
    "type": "object",
    "properties": {
        "type": {"type": "string"},
        "params": {"type": "object"},
    },
}


def _is_local_component(component_name: str) -> bool:
    return component_name.startswith(".")


def _scaffold_value_and_source_position_tree(
    filename: str, row: int, col: int
) -> ValueAndSourcePositionTree:
    return ValueAndSourcePositionTree(
        value=None,
        source_position_tree=SourcePositionTree(
            position=SourcePosition(
                filename=filename, start=LineCol(row, col), end=LineCol(row, col)
            ),
            children={},
        ),
    )


class ErrorInput(NamedTuple):
    component_name: Optional[str]
    error: ValidationError
    source_position_tree: ValueAndSourcePositionTree


@component_group.command(name="check", cls=DgClickCommand)
@click.argument("paths", nargs=-1, type=click.Path(exists=True))
@dg_global_options
@click.pass_context
def component_check_command(
    context: click.Context,
    paths: Sequence[str],
    **global_options: object,
) -> None:
    """Check component files against their schemas, showing validation errors."""
    resolved_paths = [Path(path).absolute() for path in paths]
    top_level_component_validator = Draft202012Validator(schema=COMPONENT_FILE_SCHEMA)

    cli_config = normalize_cli_config(global_options, context)
    dg_context = DgContext.for_code_location_environment(Path.cwd(), cli_config)

    validation_errors: list[ErrorInput] = []

    component_contents_by_dir = {}
    local_component_dirs = set()
    for component_dir in dg_context.components_path.iterdir():
        if resolved_paths and not any(
            path == component_dir or path in component_dir.parents for path in resolved_paths
        ):
            continue

        component_path = component_dir / "component.yaml"

        if component_path.exists():
            text = component_path.read_text()
            try:
                component_doc_tree = parse_yaml_with_source_positions(
                    text, filename=str(component_path)
                )
            except ScannerError as se:
                validation_errors.append(
                    ErrorInput(
                        None,
                        ValidationError(f"Unable to parse YAML: {se.context}, {se.problem}"),
                        _scaffold_value_and_source_position_tree(
                            filename=str(component_path),
                            row=se.problem_mark.line + 1 if se.problem_mark else 1,
                            col=se.problem_mark.column + 1 if se.problem_mark else 1,
                        ),
                    )
                )
                continue
            # First, validate the top-level structure of the component file
            # (type and params keys) before we try to validate the params themselves.
            top_level_errs = list(
                top_level_component_validator.iter_errors(component_doc_tree.value)
            )
            for err in top_level_errs:
                validation_errors.append(ErrorInput(None, err, component_doc_tree))
            if top_level_errs:
                continue

            component_contents_by_dir[component_dir] = component_doc_tree
            component_name = component_doc_tree.value.get("type")
            if _is_local_component(component_name):
                local_component_dirs.add(component_dir)

    # Fetch the local component types, if we need any local components
    component_registry = RemoteComponentRegistry.from_dg_context(
        dg_context, local_component_type_dirs=list(local_component_dirs)
    )

    for component_dir, component_doc_tree in component_contents_by_dir.items():
        component_name = component_doc_tree.value.get("type")

        try:
            json_schema = (
                component_registry.get(component_dir, component_name).component_params_schema or {}
            )

            v = Draft202012Validator(json_schema)
            for err in v.iter_errors(component_doc_tree.value["params"]):
                validation_errors.append(ErrorInput(component_name, err, component_doc_tree))
        except KeyError:
            # No matching component type found
            validation_errors.append(
                ErrorInput(
                    None,
                    ValidationError(
                        f"Unable to locate local component type '{component_name}' in {component_dir}."
                        if _is_local_component(component_name)
                        else f"No component type named '{component_name}' found."
                    ),
                    component_doc_tree,
                )
            )
    if validation_errors:
        for component_name, error, component_doc_tree in validation_errors:
            click.echo(
                error_dict_to_formatted_error(
                    component_name,
                    error,
                    source_position_tree=component_doc_tree.source_position_tree,
                    prefix=["params"] if component_name else [],
                )
            )
        context.exit(1)
    else:
        click.echo("All components validated successfully.")



import os
import signal
import subprocess
import sys
import time
from collections.abc import Iterator, Mapping, Sequence
from contextlib import contextmanager, nullcontext
from pathlib import Path
from tempfile import NamedTemporaryFile
from typing import Optional, TypeVar

import click
import psutil
import yaml

from dagster_dg.cli.global_options import dg_global_options
from dagster_dg.config import normalize_cli_config
from dagster_dg.context import DgContext
from dagster_dg.error import DgError
from dagster_dg.utils import DgClickCommand, exit_with_error, pushd

T = TypeVar("T")

_CHECK_SUBPROCESS_INTERVAL = 5


@click.command(name="dev", cls=DgClickCommand)
@click.option(
    "--code-server-log-level",
    help="Set the log level for code servers spun up by dagster services.",
    show_default=True,
    default="warning",
    type=click.Choice(["critical", "error", "warning", "info", "debug"], case_sensitive=False),
)
@click.option(
    "--log-level",
    help="Set the log level for dagster services.",
    show_default=True,
    default="info",
    type=click.Choice(["critical", "error", "warning", "info", "debug"], case_sensitive=False),
)
@click.option(
    "--log-format",
    type=click.Choice(["colored", "json", "rich"], case_sensitive=False),
    show_default=True,
    required=False,
    default="colored",
    help="Format of the logs for dagster services",
)
@click.option(
    "--port",
    "-p",
    type=int,
    help="Port to use for the Dagster webserver.",
    required=False,
)
@click.option(
    "--host",
    "-h",
    type=str,
    help="Host to use for the Dagster webserver.",
    required=False,
)
@click.option(
    "--live-data-poll-rate",
    help="Rate at which the dagster UI polls for updated asset data (in milliseconds)",
    type=int,
    default=2000,
    show_default=True,
    required=False,
)
@dg_global_options
@click.pass_context
def dev_command(
    context: click.Context,
    code_server_log_level: str,
    log_level: str,
    log_format: str,
    port: Optional[int],
    host: Optional[str],
    live_data_poll_rate: int,
    **global_options: Mapping[str, object],
) -> None:
    """Start a local deployment of your Dagster project.

    If run inside a deployment directory, this command will launch all code locations in the
    deployment. If launched inside a code location directory, it will launch only that code
    location.
    """
    cli_config = normalize_cli_config(global_options, context)
    dg_context = DgContext.from_config_file_discovery_and_cli_config(Path.cwd(), cli_config)

    forward_options = [
        *_format_forwarded_option("--code-server-log-level", code_server_log_level),
        *_format_forwarded_option("--log-level", log_level),
        *_format_forwarded_option("--log-format", log_format),
        *_format_forwarded_option("--port", port),
        *_format_forwarded_option("--host", host),
        *_format_forwarded_option("--live-data-poll-rate", live_data_poll_rate),
    ]

    # In a code location context, we can just run `dagster dev` directly, using `dagster` from the
    # code location's environment.
    if dg_context.is_code_location:
        cmd = ["uv", "run", "dagster", "dev", *forward_options]
        temp_workspace_file_cm = nullcontext()

    # In a deployment context, dg dev will construct a temporary
    # workspace file that points at all defined code locations and invoke:
    #
    #     uv tool run --with dagster-webserver dagster dev
    #
    # The `--with dagster-webserver` is necessary here to ensure that dagster-webserver is
    # installed in the isolated environment that `uv` will install `dagster` in.
    # `dagster-webserver` is not a dependency of `dagster` but is required to run the `dev`
    # command.
    elif dg_context.is_deployment:
        cmd = [
            "uv",
            "tool",
            "run",
            "--with",
            "dagster-webserver",
            "dagster",
            "dev",
            *forward_options,
        ]
        temp_workspace_file_cm = _temp_workspace_file(dg_context)
    else:
        exit_with_error("This command must be run inside a code location or deployment directory.")

    with pushd(dg_context.root_path), temp_workspace_file_cm as workspace_file:
        if workspace_file:  # only non-None deployment context
            cmd.extend(["--workspace", workspace_file])
        uv_run_dagster_dev_process = _open_subprocess(cmd)
        try:
            while True:
                time.sleep(_CHECK_SUBPROCESS_INTERVAL)
                if uv_run_dagster_dev_process.poll() is not None:
                    raise DgError(
                        f"dagster-dev process shut down unexpectedly with return code {uv_run_dagster_dev_process.returncode}."
                    )
        except KeyboardInterrupt:
            click.secho(
                "Received keyboard interrupt. Shutting down dagster-dev process.", fg="yellow"
            )
        finally:
            # For reasons not fully understood, directly interrupting the `uv run` process does not
            # work as intended. The interrupt signal is not correctly propagated to the `dagster
            # dev` process, and so that process never shuts down. Therefore, we send the signal
            # directly to the `dagster dev` process (the only child of the `uv run` process). This
            # will cause `dagster dev` to terminate which in turn will cause `uv run` to terminate.
            dagster_dev_pid = _get_child_process_pid(uv_run_dagster_dev_process)
            _interrupt_subprocess(dagster_dev_pid)

            try:
                uv_run_dagster_dev_process.wait(timeout=10)
            except subprocess.TimeoutExpired:
                click.secho("`dagster dev` did not terminate in time. Killing it.")
                uv_run_dagster_dev_process.kill()


@contextmanager
def _temp_workspace_file(dg_context: DgContext) -> Iterator[str]:
    with NamedTemporaryFile(mode="w+", delete=True) as temp_workspace_file:
        entries = []
        for location in dg_context.get_code_location_names():
            code_location_root = dg_context.get_code_location_path(location)
            loc_context = dg_context.with_root_path(code_location_root)
            entry = {
                "working_directory": str(dg_context.deployment_root_path),
                "relative_path": str(loc_context.definitions_path),
                "location_name": loc_context.code_location_name,
            }
            if loc_context.use_dg_managed_environment:
                entry["executable_path"] = str(loc_context.code_location_python_executable)
            entries.append({"python_file": entry})
        yaml.dump({"load_from": entries}, temp_workspace_file)
        temp_workspace_file.flush()
        yield temp_workspace_file.name


def _format_forwarded_option(option: str, value: object) -> list[str]:
    return [] if value is None else [option, str(value)]


def _get_child_process_pid(proc: "subprocess.Popen") -> int:
    children = psutil.Process(proc.pid).children(recursive=False)
    if len(children) != 1:
        raise ValueError(f"Expected exactly one child process, but found {len(children)}")
    return children[0].pid


# Windows subprocess termination utilities. See here for why we send CTRL_BREAK_EVENT on Windows:
# https://stefan.sofa-rockers.org/2013/08/15/handling-sub-process-hierarchies-python-linux-os-x/


def _interrupt_subprocess(pid: int) -> None:
    """Send CTRL_BREAK_EVENT on Windows, SIGINT on other platforms."""
    if sys.platform == "win32":
        os.kill(pid, signal.CTRL_BREAK_EVENT)
    else:
        os.kill(pid, signal.SIGINT)


def _open_subprocess(command: Sequence[str]) -> "subprocess.Popen":
    """Sets the correct flags to support graceful termination."""
    creationflags = 0
    if sys.platform == "win32":
        creationflags = subprocess.CREATE_NEW_PROCESS_GROUP

    return subprocess.Popen(
        command,
        creationflags=creationflags,
    )



import subprocess
from collections.abc import Iterable
from dataclasses import replace
from functools import cached_property
from pathlib import Path
from typing import Final, Optional

from typing_extensions import Self

from dagster_dg.cache import CachableDataType, DgCache, hash_paths
from dagster_dg.component import RemoteComponentRegistry
from dagster_dg.config import DgConfig, DgPartialConfig, load_dg_config_file
from dagster_dg.error import DgError
from dagster_dg.utils import (
    MISSING_DAGSTER_COMPONENTS_ERROR_MESSAGE,
    NOT_CODE_LOCATION_ERROR_MESSAGE,
    NOT_COMPONENT_LIBRARY_ERROR_MESSAGE,
    NOT_DEPLOYMENT_ERROR_MESSAGE,
    ensure_loadable_path,
    exit_with_error,
    get_path_for_module,
    get_path_for_package,
    get_uv_command_env,
    is_executable_available,
    is_package_installed,
    pushd,
)

# Deployment
_DEPLOYMENT_CODE_LOCATIONS_DIR: Final = "code_locations"

# Code location
_DEFAULT_CODE_LOCATION_COMPONENTS_LIB_SUBMODULE: Final = "lib"
_DEFAULT_CODE_LOCATION_COMPONENTS_SUBMODULE: Final = "components"


class DgContext:
    root_path: Path
    config: DgConfig
    _cache: Optional[DgCache] = None

    def __init__(self, config: DgConfig, root_path: Path):
        self.config = config
        self.root_path = root_path
        if config.disable_cache or not self.use_dg_managed_environment:
            self._cache = None
        else:
            self._cache = DgCache.from_config(config)
        self.component_registry = RemoteComponentRegistry.empty()

    @classmethod
    def for_deployment_environment(cls, path: Path, cli_config: DgPartialConfig) -> Self:
        context = cls.from_config_file_discovery_and_cli_config(path, cli_config)

        # Commands that operate on a deployment need to be run inside a deployment context.
        if not context.is_deployment:
            exit_with_error(NOT_DEPLOYMENT_ERROR_MESSAGE)
        return context

    @classmethod
    def for_code_location_environment(cls, path: Path, cli_config: DgPartialConfig) -> Self:
        context = cls.from_config_file_discovery_and_cli_config(path, cli_config)

        # Commands that operate on a code location need to be run (a) inside a code location
        # context; and (b) with dagster-components available on $PATH.
        if not context.is_code_location:
            exit_with_error(NOT_CODE_LOCATION_ERROR_MESSAGE)
        elif not is_executable_available("dagster-components"):
            exit_with_error(MISSING_DAGSTER_COMPONENTS_ERROR_MESSAGE)
        return context

    @classmethod
    def for_component_library_environment(cls, path: Path, cli_config: DgPartialConfig) -> Self:
        context = cls.from_config_file_discovery_and_cli_config(path, cli_config)

        # Commands that operate on a component library need to be run (a) inside a component
        # library context; and (b) with dagster-components available on $PATH.
        if not context.is_component_library:
            exit_with_error(NOT_COMPONENT_LIBRARY_ERROR_MESSAGE)
        elif not is_executable_available("dagster-components"):
            exit_with_error(MISSING_DAGSTER_COMPONENTS_ERROR_MESSAGE)
        return context

    @classmethod
    def for_defined_registry_environment(cls, path: Path, cli_config: DgPartialConfig) -> Self:
        context = cls.from_config_file_discovery_and_cli_config(path, cli_config)

        # Commands that access the component registry need to be run with dagster-components
        # available on $PATH.
        if not is_executable_available("dagster-components"):
            exit_with_error(MISSING_DAGSTER_COMPONENTS_ERROR_MESSAGE)
        return context

    @classmethod
    def from_config_file_discovery_and_cli_config(
        cls, path: Path, cli_config: DgPartialConfig
    ) -> Self:
        config_path = DgConfig.discover_config_file(path)
        root_path = config_path.parent if config_path else path
        base_config = DgConfig.from_config_file(config_path) if config_path else DgConfig.default()
        config = replace(base_config, **cli_config)
        return cls(config=config, root_path=root_path)

    @classmethod
    def default(cls) -> Self:
        return cls(DgConfig.default(), Path.cwd())

    # Use to derive a new context for a code location while preserving existing settings
    def with_root_path(self, root_path: Path) -> Self:
        config_path = root_path / "pyproject.toml"
        if not root_path / "pyproject.toml":
            raise DgError(f"Cannot find `pyproject.toml` at {root_path}")
        new_config = replace(self.config, **load_dg_config_file(config_path))
        return self.__class__(config=new_config, root_path=root_path)

    # ########################
    # ##### CACHE METHODS
    # ########################

    @property
    def cache(self) -> DgCache:
        if not self._cache:
            raise DgError("Cache is disabled")
        return self._cache

    @property
    def has_cache(self) -> bool:
        return self._cache is not None

    def get_cache_key(self, data_type: CachableDataType) -> tuple[str, str, str]:
        path_parts = [str(part) for part in self.root_path.parts if part != "/"]
        paths_to_hash = [
            self.root_path / "uv.lock",
            *([self.components_lib_path] if self.is_component_library else []),
        ]
        env_hash = hash_paths(paths_to_hash)
        return ("_".join(path_parts), env_hash, data_type)

    def get_cache_key_for_local_components(self, path: Path) -> tuple[str, str, str]:
        env_hash = hash_paths([path], includes=["*.py"])
        path_parts = [str(part) for part in path.parts if part != "/"]
        return ("_".join(path_parts), env_hash, "local_component_registry")

    # ########################
    # ##### DEPLOYMENT METHODS
    # ########################

    @property
    def is_deployment(self) -> bool:
        return self.config.is_deployment

    @cached_property
    def deployment_root_path(self) -> Path:
        if not self.is_deployment:
            raise DgError("`deployment_root_path` is only available in a deployment context")
        deployment_config_path = DgConfig.discover_config_file(
            self.root_path, lambda x: x.get("is_deployment", False)
        )
        if not deployment_config_path:
            raise DgError("Cannot find deployment configuration file")
        return deployment_config_path.parent

    def has_code_location(self, name: str) -> bool:
        if not self.is_deployment:
            raise DgError(
                "`deployment_has_code_location` is only available in a deployment context"
            )
        return (self.deployment_root_path / _DEPLOYMENT_CODE_LOCATIONS_DIR / name).is_dir()

    @property
    def code_location_root_path(self) -> Path:
        if not self.is_deployment:
            raise DgError(
                "`deployment_code_location_root_path` is only available in a deployment context"
            )
        return self.deployment_root_path / _DEPLOYMENT_CODE_LOCATIONS_DIR

    def get_code_location_names(self) -> Iterable[str]:
        return [loc.name for loc in sorted(self.code_location_root_path.iterdir())]

    def get_code_location_path(self, name: str) -> Path:
        return self.code_location_root_path / name

    def get_code_location_root_module(self, name: str) -> Path:
        return self.code_location_root_path / name

    # ########################
    # ##### GENERAL PYTHON PACKAGE METHODS
    # ########################

    @property
    def root_package_name(self) -> str:
        return self.config.root_package or self.root_path.name.replace("-", "_")

    # ########################
    # ##### CODE LOCATION METHODS
    # ########################

    @property
    def is_code_location(self) -> bool:
        return self.config.is_code_location

    @property
    def code_location_name(self) -> str:
        if not self.is_code_location:
            raise DgError("`code_location_name` is only available in a code location context")
        return self.root_path.name

    @property
    def code_location_python_executable(self) -> Path:
        if not self.is_code_location:
            raise DgError(
                "`code_location_python_executable` is only available in a code location context"
            )
        return self.root_path / ".venv" / "bin" / "python"

    @cached_property
    def components_package_name(self) -> str:
        if not self.is_code_location:
            raise DgError("`components_package_name` is only available in a code location context")
        return (
            self.config.component_package
            or f"{self.root_package_name}.{_DEFAULT_CODE_LOCATION_COMPONENTS_SUBMODULE}"
        )

    @cached_property
    def components_path(self) -> Path:
        if not self.is_code_location:
            raise DgError("`components_path` is only available in a code location context")
        with ensure_loadable_path(self.root_path):
            if not is_package_installed(self.components_package_name):
                raise DgError(
                    f"Components package `{self.components_package_name}` is not installed in the current environment."
                )
            return Path(get_path_for_package(self.components_package_name))

    def get_component_names(self) -> Iterable[str]:
        return [str(instance_path.name) for instance_path in self.components_path.iterdir()]

    def has_component(self, name: str) -> bool:
        return (self.components_path / name).is_dir()

    @property
    def definitions_package_name(self) -> str:
        if not self.is_code_location:
            raise DgError("`definitions_package_name` is only available in a code location context")
        return f"{self.root_package_name}.definitions"

    @cached_property
    def definitions_path(self) -> Path:
        with ensure_loadable_path(self.root_path):
            if not is_package_installed(self.definitions_package_name):
                raise DgError(
                    f"Definitions package `{self.definitions_package_name}` is not installed in the current environment."
                )
            return Path(get_path_for_module(self.definitions_package_name))

    # ########################
    # ##### COMPONENT LIBRARY METHODS
    # ########################

    @property
    def is_component_library(self) -> bool:
        return self.config.is_component_lib

    @cached_property
    def components_lib_package_name(self) -> str:
        if not self.is_component_library:
            raise DgError(
                "`components_lib_package_name` is only available in a component library context"
            )
        return (
            self.config.component_lib_package
            or f"{self.root_package_name}.{_DEFAULT_CODE_LOCATION_COMPONENTS_LIB_SUBMODULE}"
        )

    @cached_property
    def components_lib_path(self) -> Path:
        if not self.is_component_library:
            raise DgError("`components_lib_path` is only available in a component library context")
        with ensure_loadable_path(self.root_path):
            if not is_package_installed(self.components_lib_package_name):
                raise DgError(
                    f"Components lib package `{self.components_lib_package_name}` is not installed in the current environment."
                )
            return Path(get_path_for_package(self.components_lib_package_name))

    # ########################
    # ##### HELPERS
    # ########################

    def external_components_command(self, command: list[str]) -> str:
        if self.use_dg_managed_environment:
            code_location_command_prefix = ["uv", "run", "dagster-components"]
            env = get_uv_command_env()
        else:
            code_location_command_prefix = ["dagster-components"]
            env = None
        full_command = [
            *code_location_command_prefix,
            *(
                ["--builtin-component-lib", self.config.builtin_component_lib]
                if self.config.builtin_component_lib
                else []
            ),
            *command,
        ]
        with pushd(self.root_path):
            result = subprocess.run(full_command, stdout=subprocess.PIPE, env=env, check=True)
            return result.stdout.decode("utf-8")

    def ensure_uv_lock(self, path: Optional[Path] = None) -> None:
        path = path or self.root_path
        with pushd(path):
            if not (path / "uv.lock").exists():
                subprocess.run(["uv", "sync"], check=True, env=get_uv_command_env())

    @property
    def use_dg_managed_environment(self) -> bool:
        return self.config.use_dg_managed_environment and self.is_code_location

    def validate_deployment_command_environment(self) -> None:
        """Commands that operate on a deployment need to be run inside a deployment context."""
        if not self.is_deployment:
            exit_with_error(NOT_DEPLOYMENT_ERROR_MESSAGE)

    def validate_registry_command_environment(self) -> None:
        """Commands that access the component registry need to be run with dagster-components
        available on $PATH.
        """
        if not is_executable_available("dagster-components"):
            exit_with_error(MISSING_DAGSTER_COMPONENTS_ERROR_MESSAGE)



import contextlib
import importlib.util
import json
import os
import posixpath
import re
import shutil
import sys
import textwrap
from collections.abc import Iterator, Mapping, Sequence
from fnmatch import fnmatch
from importlib.machinery import ModuleSpec
from pathlib import Path
from typing import Any, Optional, TypeVar, Union

import click
import jinja2
from typer.rich_utils import rich_format_help
from typing_extensions import Never, TypeAlias

from dagster_dg.error import DgError
from dagster_dg.version import __version__ as dagster_version

# There is some weirdness concerning the availabilty of hashlib.HASH between different Python
# versions, so for nowe we avoid trying to import it and just alias the type to Any.
Hash: TypeAlias = Any

CLI_CONFIG_KEY = "config"


# Temporarily places a path at the front of sys.path, ensuring that any modules in that path are
# importable.
@contextlib.contextmanager
def ensure_loadable_path(path: Path) -> Iterator[None]:
    orig_path = sys.path.copy()
    sys.path.insert(0, str(path))
    try:
        yield
    finally:
        sys.path = orig_path


def is_package_installed(package_name: str) -> bool:
    try:
        return bool(importlib.util.find_spec(package_name))
    except ModuleNotFoundError:
        return False


def _get_spec_for_module(module_name: str) -> ModuleSpec:
    spec = importlib.util.find_spec(module_name)
    if not spec:
        raise DgError(f"Cannot find module: {module_name}")
    return spec


def get_path_for_module(module_name: str) -> str:
    spec = _get_spec_for_module(module_name)
    file_path = spec.origin
    if not file_path:
        raise DgError(f"Cannot find file path for module: {module_name}")
    return file_path


def get_path_for_package(package_name: str) -> str:
    spec = _get_spec_for_module(package_name)
    submodule_search_locations = spec.submodule_search_locations
    if not submodule_search_locations:
        raise DgError(f"Package does not have any locations for submodules: {package_name}")
    return submodule_search_locations[0]


def is_valid_json(value: str) -> bool:
    try:
        json.loads(value)
        return True
    except json.JSONDecodeError:
        return False


def is_executable_available(command: str) -> bool:
    return bool(shutil.which(command))


# uv commands should be executed in an environment with no pre-existing VIRTUAL_ENV set. If this
# variable is set (common during development) and does not match the venv resolved by uv, it prints
# undesireable warnings.
def get_uv_command_env() -> Mapping[str, str]:
    return {k: v for k, v in os.environ.items() if not k == "VIRTUAL_ENV"}


def discover_git_root(path: Path) -> Path:
    while path != path.parent:
        if (path / ".git").exists():
            return path
        path = path.parent
    raise ValueError("Could not find git root")


@contextlib.contextmanager
def pushd(path: Union[str, Path]) -> Iterator[None]:
    old_cwd = os.getcwd()
    os.chdir(path)
    try:
        yield
    finally:
        os.chdir(old_cwd)


def camelcase(string: str) -> str:
    string = re.sub(r"^[\-_\.]", "", str(string))
    if not string:
        return string
    return str(string[0]).upper() + re.sub(
        r"[\-_\.\s]([a-z])", lambda matched: str(matched.group(1)).upper(), string[1:]
    )


def snakecase(string: str) -> str:
    # Add an underscore before capital letters and lower the case
    string = re.sub(r"(?<!^)(?=[A-Z])", "_", string).lower()
    # Replace any non-alphanumeric characters with underscores
    string = re.sub(r"[^a-z0-9_]", "_", string)
    return string


_HELP_OUTPUT_GROUP_ATTR = "rich_help_panel"


# This is a private API of typer, but it is the only way I can see to get our global options grouped
# differently and it works well.
def set_option_help_output_group(param: click.Parameter, group: str) -> None:
    """Sets the help output group for a click parameter."""
    setattr(param, "rich_help_panel", group)


DEFAULT_FILE_EXCLUDE_PATTERNS: list[str] = [
    "__pycache__",
    ".pytest_cache",
    "*.egg-info",
    "*.cpython-*",
    ".DS_Store",
    ".ruff_cache",
    "tox.ini",
    ".gitkeep",  # dummy file that allows empty directories to be checked into git
]

PROJECT_NAME_PLACEHOLDER = "PROJECT_NAME_PLACEHOLDER"


# Copied from dagster._generate.generate
def scaffold_subtree(
    path: Path,
    excludes: Optional[list[str]] = None,
    name_placeholder: str = PROJECT_NAME_PLACEHOLDER,
    templates_path: str = PROJECT_NAME_PLACEHOLDER,
    project_name: Optional[str] = None,
    **other_template_vars: Any,
):
    """Renders templates for Dagster project."""
    excludes = (
        DEFAULT_FILE_EXCLUDE_PATTERNS if not excludes else DEFAULT_FILE_EXCLUDE_PATTERNS + excludes
    )

    normalized_path = os.path.normpath(path)
    project_name = project_name or os.path.basename(normalized_path).replace("-", "_")
    if not os.path.exists(normalized_path):
        os.mkdir(normalized_path)

    project_template_path = os.path.join(os.path.dirname(__file__), "templates", templates_path)
    loader = jinja2.FileSystemLoader(searchpath=project_template_path)
    env = jinja2.Environment(loader=loader)

    # merge custom skip_files with the default list
    for root, dirs, files in os.walk(project_template_path):
        # For each subdirectory in the source template, create a subdirectory in the destination.
        for dirname in dirs:
            src_dir_path = os.path.join(root, dirname)
            if _should_skip_file(src_dir_path, excludes):
                continue

            src_relative_dir_path = os.path.relpath(src_dir_path, project_template_path)
            dst_relative_dir_path = src_relative_dir_path.replace(
                name_placeholder,
                project_name,
                1,
            )
            dst_dir_path = os.path.join(normalized_path, dst_relative_dir_path)

            os.mkdir(dst_dir_path)

        # For each file in the source template, render a file in the destination.
        for filename in files:
            src_file_path = os.path.join(root, filename)
            if _should_skip_file(src_file_path, excludes):
                continue

            src_relative_file_path = os.path.relpath(src_file_path, project_template_path)
            dst_relative_file_path = src_relative_file_path.replace(
                name_placeholder,
                project_name,
                1,
            )
            dst_file_path = os.path.join(normalized_path, dst_relative_file_path)

            if dst_file_path.endswith(".jinja"):
                dst_file_path = dst_file_path[: -len(".jinja")]

            with open(dst_file_path, "w", encoding="utf8") as f:
                # Jinja template names must use the POSIX path separator "/".
                template_name = src_relative_file_path.replace(os.sep, posixpath.sep)
                template: jinja2.environment.Template = env.get_template(name=template_name)
                f.write(
                    template.render(
                        repo_name=project_name,  # deprecated
                        dagster_version=dagster_version,
                        project_name=project_name,
                        **other_template_vars,
                    )
                )
                f.write("\n")

    click.echo(f"Scaffolded files for Dagster project in {path}.")


def _should_skip_file(path: str, excludes: list[str] = DEFAULT_FILE_EXCLUDE_PATTERNS):
    """Given a file path `path` in a source template, returns whether or not the file should be skipped
    when generating destination files.

    Technically, `path` could also be a directory path that should be skipped.
    """
    for pattern in excludes:
        if pattern.lower() in path.lower():
            return True

    return False


def ensure_dagster_dg_tests_import() -> None:
    from dagster_dg import __file__ as dagster_dg_init_py

    dagster_dg_package_root = (Path(dagster_dg_init_py) / ".." / "..").resolve()
    assert (
        dagster_dg_package_root / "dagster_dg_tests"
    ).exists(), "Could not find dagster_dg_tests where expected"
    sys.path.append(dagster_dg_package_root.as_posix())


def hash_directory_metadata(
    hasher: Hash,
    path: Union[str, Path],
    includes: Optional[Sequence[str]],
    excludes: Sequence[str],
) -> None:
    for root, dirs, files in os.walk(path):
        for name in dirs + files:
            if any(fnmatch(name, pattern) for pattern in excludes):
                continue
            if includes and not any(fnmatch(name, pattern) for pattern in includes):
                continue
            filepath = os.path.join(root, name)
            hash_file_metadata(hasher, filepath)


def hash_file_metadata(hasher: Hash, path: Union[str, Path]) -> None:
    stat = os.stat(path=path)
    hasher.update(str(path).encode())
    hasher.update(str(stat.st_mtime).encode())  # Last modified time
    hasher.update(str(stat.st_size).encode())  # File size


T = TypeVar("T")


def not_none(value: Optional[T]) -> T:
    if value is None:
        raise DgError("Expected non-none value.")
    return value


def exit_with_error(error_message: str) -> Never:
    click.echo(click.style(error_message, fg="red"))
    sys.exit(1)


def _format_error_message(message: str) -> str:
    # width=10000 unwraps any hardwrapping
    return textwrap.fill(message, width=10000)


NOT_DEPLOYMENT_ERROR_MESSAGE = _format_error_message("""
This command must be run inside a Dagster deployment directory. Ensure that there is a
`pyproject.toml` file with `tool.dg.is_deployment = true` set in the root deployment directory.
""")


NOT_CODE_LOCATION_ERROR_MESSAGE = _format_error_message("""
This command must be run inside a Dagster code location directory. Ensure that the nearest
pyproject.toml has `tool.dg.is_code_location = true` set.
""")

NOT_COMPONENT_LIBRARY_ERROR_MESSAGE = _format_error_message("""
This command must be run inside a Dagster component library directory. Ensure that the nearest
pyproject.toml has `tool.dg.is_component_lib = true` set.
""")


MISSING_DAGSTER_COMPONENTS_ERROR_MESSAGE = _format_error_message("""
Could not find the `dagster-components` executable on the system path.

The `dagster-components` executable is installed with the `dagster-components` PyPI package and is
necessary for `dg` to interface with Python environments containing Dagster definitions.
`dagster-components` is installed by default when a code location is scaffolded by `dg`. However, if
you are using `dg` in a non-managed environment (either outside of a code location or using the
`--no-use-dg-managed-environment` flag), you need to independently ensure `dagster-components` is
installed.
""")

# ########################
# ##### CUSTOM CLICK SUBCLASSES
# ########################

# Here we subclass click.Command and click.Group to customize the help output. We do this in order
# to visually separate global from command-specific options. The form of the output can be seen in
# dagster_dg_tests.test_custom_help_format.


class DgClickHelpMixin:
    def format_help(self, context: click.Context, formatter: click.HelpFormatter):
        """Customizes the help to include hierarchical usage."""
        if not isinstance(self, click.Command):
            raise ValueError("This mixin is only intended for use with click.Command instances.")

        # We use typer's rich_format_help to render our help output, despite the fact that we are
        # not using typer elsewhere in the app. Global options are separated from command-specific
        # options by setting the `rich_help_panel` attribute to "Global options" on our global options.
        rich_format_help(obj=self, ctx=context, markup_mode="rich")


class DgClickCommand(DgClickHelpMixin, click.Command): ...


class DgClickGroup(DgClickHelpMixin, click.Group): ...


# ########################
# ##### JSON SCHEMA
# ########################

_JSON_SCHEMA_TYPE_TO_CLICK_TYPE = {"string": str, "integer": int, "number": float, "boolean": bool}


def json_schema_property_to_click_option(
    key: str, field_info: Mapping[str, Any], required: bool
) -> click.Option:
    field_type = field_info.get("type", "string")
    option_name = f"--{key.replace('_', '-')}"
    # Handle object type fields as JSON strings
    if field_type == "object":
        option_type = str  # JSON string input
        help_text = f"{key} (JSON string)"
        callback = parse_json_option

    # Handle other basic types
    else:
        option_type = _JSON_SCHEMA_TYPE_TO_CLICK_TYPE[field_type]
        help_text = key
        callback = None

    return click.Option(
        [option_name],
        type=option_type,
        required=required,
        help=help_text,
        callback=callback,
    )


def parse_json_option(context: click.Context, param: click.Option, value: str):
    """Callback to parse JSON string options into Python objects."""
    if value:
        try:
            return json.loads(value)
        except json.JSONDecodeError:
            raise click.BadParameter(f"Invalid JSON string for '{param.name}'.")
    return value



import json
from collections.abc import Mapping
from pathlib import Path
from typing import Any

import click
from rich.console import Console
from rich.table import Table

from dagster_dg.cli.global_options import dg_global_options
from dagster_dg.component import RemoteComponentRegistry
from dagster_dg.config import normalize_cli_config
from dagster_dg.context import DgContext
from dagster_dg.docs import markdown_for_component_type, render_markdown_in_browser
from dagster_dg.scaffold import scaffold_component_type
from dagster_dg.utils import DgClickCommand, DgClickGroup, exit_with_error


@click.group(name="component-type", cls=DgClickGroup)
def component_type_group():
    """Commands for operating on components types."""


# ########################
# ##### SCAFFOLD
# ########################


@component_type_group.command(name="scaffold", cls=DgClickCommand)
@click.argument("name", type=str)
@dg_global_options
@click.pass_context
def component_type_scaffold_command(
    context: click.Context, name: str, **global_options: object
) -> None:
    """Scaffold of a custom Dagster component type.

    This command must be run inside a Dagster code location directory. The component type scaffold
    will be placed in submodule `<code_location_name>.lib.<name>`.
    """
    cli_config = normalize_cli_config(global_options, context)
    dg_context = DgContext.for_component_library_environment(Path.cwd(), cli_config)
    registry = RemoteComponentRegistry.from_dg_context(dg_context)
    full_component_name = f"{dg_context.root_package_name}.{name}"
    if registry.has_global(full_component_name):
        exit_with_error(f"A component type named `{name}` already exists.")

    scaffold_component_type(dg_context, name)


# ########################
# ##### DOCS
# ########################


@component_type_group.command(name="docs", cls=DgClickCommand)
@click.argument("component_type", type=str)
@dg_global_options
@click.pass_context
def component_type_docs_command(
    context: click.Context,
    component_type: str,
    **global_options: object,
) -> None:
    """Get detailed information on a registered Dagster component type."""
    cli_config = normalize_cli_config(global_options, context)
    dg_context = DgContext.for_defined_registry_environment(Path.cwd(), cli_config)
    registry = RemoteComponentRegistry.from_dg_context(dg_context)
    if not registry.has_global(component_type):
        exit_with_error(f"No component type `{component_type}` could be resolved.")

    render_markdown_in_browser(markdown_for_component_type(registry.get_global(component_type)))


# ########################
# ##### INFO
# ########################


@component_type_group.command(name="info", cls=DgClickCommand)
@click.argument("component_type", type=str)
@click.option("--description", is_flag=True, default=False)
@click.option("--scaffold-params-schema", is_flag=True, default=False)
@click.option("--component-params-schema", is_flag=True, default=False)
@dg_global_options
@click.pass_context
def component_type_info_command(
    context: click.Context,
    component_type: str,
    description: bool,
    scaffold_params_schema: bool,
    component_params_schema: bool,
    **global_options: object,
) -> None:
    """Get detailed information on a registered Dagster component type."""
    cli_config = normalize_cli_config(global_options, context)
    dg_context = DgContext.for_defined_registry_environment(Path.cwd(), cli_config)
    registry = RemoteComponentRegistry.from_dg_context(dg_context)
    if not registry.has_global(component_type):
        exit_with_error(f"No component type `{component_type}` could be resolved.")
    elif sum([description, scaffold_params_schema, component_params_schema]) > 1:
        exit_with_error(
            "Only one of --description, --scaffold-params-schema, and --component-params-schema can be specified."
        )

    component_type_metadata = registry.get_global(component_type)

    if description:
        if component_type_metadata.description:
            click.echo(component_type_metadata.description)
        else:
            click.echo("No description available.")
    elif scaffold_params_schema:
        if component_type_metadata.scaffold_params_schema:
            click.echo(_serialize_json_schema(component_type_metadata.scaffold_params_schema))
        else:
            click.echo("No scaffold params schema defined.")
    elif component_params_schema:
        if component_type_metadata.component_params_schema:
            click.echo(_serialize_json_schema(component_type_metadata.component_params_schema))
        else:
            click.echo("No component params schema defined.")

    # print all available metadata
    else:
        click.echo(component_type)
        if component_type_metadata.description:
            click.echo("\nDescription:\n")
            click.echo(component_type_metadata.description)
        if component_type_metadata.scaffold_params_schema:
            click.echo("\nScaffold params schema:\n")
            click.echo(_serialize_json_schema(component_type_metadata.scaffold_params_schema))
        if component_type_metadata.component_params_schema:
            click.echo("\nComponent params schema:\n")
            click.echo(_serialize_json_schema(component_type_metadata.component_params_schema))


def _serialize_json_schema(schema: Mapping[str, Any]) -> str:
    return json.dumps(schema, indent=4)


# ########################
# ##### LIST
# ########################


@component_type_group.command(name="list", cls=DgClickCommand)
@dg_global_options
@click.pass_context
def component_type_list(context: click.Context, **global_options: object) -> None:
    """List registered Dagster components in the current code location environment."""
    cli_config = normalize_cli_config(global_options, context)
    dg_context = DgContext.for_defined_registry_environment(Path.cwd(), cli_config)
    registry = RemoteComponentRegistry.from_dg_context(dg_context)

    table = Table(border_style="dim")
    table.add_column("Component Type", style="bold cyan", no_wrap=True)
    table.add_column("Summary")
    for key in sorted(registry.global_keys()):
        table.add_row(key, registry.get_global(key).summary)
    console = Console()
    console.print(table)



import sys
from collections.abc import Mapping
from dataclasses import dataclass, replace
from pathlib import Path
from typing import Any, Callable, Optional, TypedDict, TypeVar, cast

import click
import tomli
from click.core import ParameterSource

from dagster_dg.error import DgError, DgValidationError

T = TypeVar("T")

DEFAULT_BUILTIN_COMPONENT_LIB = "dagster_components"


def _get_default_cache_dir() -> Path:
    if sys.platform == "win32":
        return Path.home() / "AppData" / "dg" / "cache"
    elif sys.platform == "darwin":
        return Path.home() / "Library" / "Caches" / "dg"
    else:
        return Path.home() / ".cache" / "dg"


DEFAULT_CACHE_DIR = _get_default_cache_dir()


@dataclass
class DgConfig:
    """Global configuration for Dg.

    Attributes:
        disable_cache (bool): If True, disable caching. Defaults to False.
        cache_dir (Optional[str]): The directory to use for caching. If None, the default cache will
            be used.
        verbose (bool): If True, log debug information.
        builitin_component_lib (str): The name of the builtin component library to load.
        use_dg_managed_environment (bool): If True, `dg` will build and manage a virtual environment
            using `uv`. Note that disabling the managed enviroment will also disable caching.
    """

    disable_cache: bool = False
    cache_dir: Path = DEFAULT_CACHE_DIR
    verbose: bool = False
    builtin_component_lib: str = DEFAULT_BUILTIN_COMPONENT_LIB
    use_dg_managed_environment: bool = True
    is_component_lib: bool = False
    is_code_location: bool = False
    is_deployment: bool = False
    root_package: Optional[str] = None
    component_package: Optional[str] = None
    component_lib_package: Optional[str] = None

    @classmethod
    def discover_config_file(
        cls,
        path: Path,
        predicate: Optional[Callable[[Mapping[str, Any]], bool]] = None,
    ) -> Optional[Path]:
        current_path = path.absolute()
        while True:
            config_path = current_path / "pyproject.toml"
            if config_path.exists() and is_dg_config_file(config_path, predicate):
                return config_path
            if current_path == current_path.parent:  # root
                return
            current_path = current_path.parent

    @classmethod
    def from_config_file(cls, path: Path) -> "DgConfig":
        current_config = load_dg_config_file(path)
        current_directory_path = path.parent
        while "extend" in current_config:
            extend_path = current_directory_path / current_config["extend"]
            if not is_dg_config_file(extend_path):
                raise DgValidationError(
                    "Config file {extend_path} was specified in `extend` field but does not contain a `tool.dg` section."
                )
            extend_config = load_dg_config_file(extend_path)
            current_config = cast(DgPartialFileConfig, {**extend_config, **current_config})
            current_directory_path = extend_path.parent
        return replace(DgConfig.default(), **current_config)

    @classmethod
    def default(cls) -> "DgConfig":
        return cls()


class DgPartialConfig(TypedDict, total=False):
    disable_cache: bool
    cache_dir: Path
    verbose: bool
    builtin_component_lib: str
    use_dg_managed_environment: bool
    component_package: str
    component_lib_package: str
    is_code_location: bool
    is_component_lib: bool
    is_deployment: bool


def _normalize_dg_partial_config(raw_dict: Mapping[str, object]) -> DgPartialConfig:
    config = {**raw_dict}  # copy so we can modify it
    if "disable_cache" in config and not isinstance(config["disable_cache"], bool):
        raise DgValidationError("`disable_cache` must be a boolean.")
    if "cache_dir" in config:
        if not isinstance(config["cache_dir"], (Path, str)):
            raise DgValidationError("`cache_dir` must be a string.")
        elif isinstance(config["cache_dir"], str):
            config["cache_dir"] = Path(config["cache_dir"])
    if "verbose" in config and not isinstance(config["verbose"], bool):
        raise DgValidationError("`verbose` must be a boolean.")
    if "builtin_component_lib" in config and not isinstance(config["builtin_component_lib"], str):
        raise DgValidationError("`builtin_component_lib` must be a string.")
    if "use_dg_managed_environment" in config and not isinstance(
        config["use_dg_managed_environment"], bool
    ):
        raise DgValidationError("`use_dg_managed_environment` must be a boolean.")
    if "component_package" in config and not isinstance(config["component_package"], str):
        raise DgValidationError("`component_package` must be a string.")
    if "component_lib_package" in config and not isinstance(config["component_lib_package"], str):
        raise DgValidationError("`component_lib_package` must be a string.")
    if "is_code_location" in config and not isinstance(config["is_code_location"], bool):
        raise DgValidationError("`is_code_location` must be a boolean.")
    if "is_component_lib" in config and not isinstance(config["is_component_lib"], bool):
        raise DgValidationError("`is_component_lib` must be a boolean.")
    if "is_deployment" in config and not isinstance(config["is_deployment"], bool):
        raise DgValidationError("`is_deployment` must be a boolean.")

    if unrecognized_keys := [k for k in config.keys() if k not in DgPartialConfig.__annotations__]:
        raise DgValidationError(f"Unrecognized fields in configuration: {unrecognized_keys}")
    return cast(DgPartialConfig, config)


# ########################
# ##### CONFIG CLI OPTIONS
# ########################


def normalize_cli_config(
    cli_options: Mapping[str, object], cli_context: click.Context
) -> DgPartialConfig:
    # Remove any options that weren't explicitly provided.
    filtered_options = {
        key: value
        for key, value in cli_options.items()
        if cli_context.get_parameter_source(key) != ParameterSource.DEFAULT
    }
    return _normalize_dg_partial_config(filtered_options)


_CLI_CONTEXT_CONFIG_KEY = "config"


def set_config_on_cli_context(cli_context: click.Context, config: DgPartialConfig) -> None:
    cli_context.ensure_object(dict)
    cli_context.obj[_CLI_CONTEXT_CONFIG_KEY] = config


def has_config_on_cli_context(cli_context: click.Context) -> bool:
    return _CLI_CONTEXT_CONFIG_KEY in cli_context.ensure_object(dict)


def get_config_from_cli_context(cli_context: click.Context) -> DgPartialConfig:
    cli_context.ensure_object(dict)
    return cli_context.obj[_CLI_CONTEXT_CONFIG_KEY]


# ########################
# ##### CONFIG FILE LOADING
# ########################


class DgPartialFileConfig(DgPartialConfig, total=False):
    extend: str


def _validate_dg_partial_file_config(
    raw_dict: Mapping[str, object], file_path: Path
) -> DgPartialFileConfig:
    if "extend" in raw_dict:
        if not isinstance(raw_dict["extend"], str):
            _raise_file_config_validation_error("`extend` must be a string.", file_path)
        elif not (file_path.parent / raw_dict["extend"]).exists():
            _raise_file_config_validation_error(
                "Config specifies `extend` setting to non-existent file: {raw_dict['extend']}",
                file_path,
            )
    try:
        _normalize_dg_partial_config({k: v for k, v in raw_dict.items() if k not in ["extend"]})
    except DgValidationError as e:
        _raise_file_config_validation_error(str(e), file_path)
    return cast(DgPartialFileConfig, raw_dict)


def is_dg_config_file(
    path: Path, predicate: Optional[Callable[[Mapping[str, Any]], bool]] = None
) -> bool:
    toml = tomli.loads(path.read_text())
    return "dg" in toml.get("tool", {}) and (predicate(toml["tool"]["dg"]) if predicate else True)


def load_dg_config_file(path: Path) -> DgPartialFileConfig:
    toml = tomli.loads(path.read_text())
    return _validate_dg_partial_file_config(toml["tool"]["dg"], path)


def _raise_file_config_validation_error(message: str, file_path: Path) -> None:
    raise DgError(f"Error in configuration file {file_path}: {message}")



from pathlib import Path

import click

from dagster_dg.cli.code_location import code_location_group
from dagster_dg.cli.component import component_group
from dagster_dg.cli.component_type import component_type_group
from dagster_dg.cli.deployment import deployment_group
from dagster_dg.cli.dev import dev_command
from dagster_dg.cli.global_options import dg_global_options
from dagster_dg.component import RemoteComponentRegistry
from dagster_dg.config import normalize_cli_config
from dagster_dg.context import DgContext
from dagster_dg.utils import DgClickGroup, exit_with_error
from dagster_dg.version import __version__

DG_CLI_MAX_OUTPUT_WIDTH = 120


def create_dg_cli():
    @click.group(
        name="dg",
        commands={
            "code-location": code_location_group,
            "component": component_group,
            "component-type": component_type_group,
            "deployment": deployment_group,
            "dev": dev_command,
        },
        context_settings={
            "max_content_width": DG_CLI_MAX_OUTPUT_WIDTH,
            "help_option_names": ["-h", "--help"],
        },
        invoke_without_command=True,
        cls=DgClickGroup,
    )
    @dg_global_options
    @click.option(
        "--clear-cache",
        is_flag=True,
        help="Clear the cache.",
        default=False,
    )
    @click.option(
        "--rebuild-component-registry",
        is_flag=True,
        help=(
            "Recompute and cache the set of available component types for the current environment."
            " Note that this also happens automatically whenever the cache is detected to be stale."
        ),
        default=False,
    )
    @click.option(
        "--install-completion",
        is_flag=True,
        help="Automatically detect your shell and install a completion script for the `dg` command. This will append to your shell startup file.",
        default=False,
    )
    @click.version_option(__version__, "--version", "-v")
    @click.pass_context
    def group(
        context: click.Context,
        install_completion: bool,
        clear_cache: bool,
        rebuild_component_registry: bool,
        **global_options: object,
    ):
        """CLI for working with Dagster components."""
        if install_completion:
            import dagster_dg.completion

            dagster_dg.completion.install_completion(context)
            context.exit(0)
        elif clear_cache and rebuild_component_registry:
            exit_with_error("Cannot specify both --clear-cache and --rebuild-component-registry.")
        elif clear_cache:
            cli_config = normalize_cli_config(global_options, context)
            dg_context = DgContext.from_config_file_discovery_and_cli_config(Path.cwd(), cli_config)
            dg_context.cache.clear_all()
            if context.invoked_subcommand is None:
                context.exit(0)
        elif rebuild_component_registry:
            cli_config = normalize_cli_config(global_options, context)
            dg_context = DgContext.from_config_file_discovery_and_cli_config(Path.cwd(), cli_config)
            if context.invoked_subcommand is not None:
                exit_with_error("Cannot specify --rebuild-component-registry with a subcommand.")
            _rebuild_component_registry(dg_context)
        elif context.invoked_subcommand is None:
            click.echo(context.get_help())
            context.exit(0)

    return group


def _rebuild_component_registry(dg_context: DgContext):
    if not dg_context.is_code_location:
        exit_with_error("This command must be run inside a Dagster code location directory.")
    elif not dg_context.has_cache:
        exit_with_error("Cache is disabled. This command cannot be run without a cache.")
    elif not dg_context.config.use_dg_managed_environment:
        exit_with_error(
            "Cannot rebuild the component registry with environment management disabled."
        )
    dg_context.ensure_uv_lock()
    key = dg_context.get_cache_key("component_registry_data")
    dg_context.cache.clear_key(key)
    # This will trigger a rebuild of the component registry
    RemoteComponentRegistry.from_dg_context(dg_context)


ENV_PREFIX = "DAGSTER_DG"
cli = create_dg_cli()


def main():
    cli(auto_envvar_prefix=ENV_PREFIX)



import json
import os
from collections.abc import Mapping, Sequence
from pathlib import Path
from typing import Any, Optional

import click

from dagster_dg.component import RemoteComponentRegistry
from dagster_dg.context import DgContext
from dagster_dg.utils import camelcase, scaffold_subtree

# ########################
# ##### DEPLOYMENT
# ########################


def scaffold_deployment(path: Path, dg_context: DgContext) -> None:
    click.echo(f"Creating a Dagster deployment at {path}.")

    scaffold_subtree(
        path=path,
        name_placeholder="DEPLOYMENT_NAME_PLACEHOLDER",
        templates_path=os.path.join(
            os.path.dirname(__file__), "templates", "DEPLOYMENT_NAME_PLACEHOLDER"
        ),
    )


# ########################
# ##### CODE LOCATION
# ########################

# Despite the fact that editable dependencies are resolved through tool.uv.sources, we need to set
# the dependencies themselves differently depending on whether we are using editable dagster or
# not. This is because `tool.uv.sources` only seems to apply to direct dependencies of the package,
# so any 2+-order Dagster dependency of our package needs to be listed as a direct dependency in the
# editable case.
EDITABLE_DAGSTER_DEPENDENCIES = (
    "dagster",
    "dagster-pipes",
    "dagster-components",
)
EDITABLE_DAGSTER_DEV_DEPENDENCIES = ("dagster-webserver", "dagster-graphql")
PYPI_DAGSTER_DEPENDENCIES = ("dagster-components",)
PYPI_DAGSTER_DEV_DEPENDENCIES = ("dagster-webserver",)


def get_pyproject_toml_dependencies(use_editable_dagster: bool) -> str:
    deps = EDITABLE_DAGSTER_DEPENDENCIES if use_editable_dagster else PYPI_DAGSTER_DEPENDENCIES
    return "\n".join(
        [
            "dependencies = [",
            *[f'    "{dep}",' for dep in deps],
            "]",
        ]
    )


def get_pyproject_toml_dev_dependencies(use_editable_dagster: bool) -> str:
    deps = (
        EDITABLE_DAGSTER_DEV_DEPENDENCIES if use_editable_dagster else PYPI_DAGSTER_DEV_DEPENDENCIES
    )
    return "\n".join(
        [
            "dev = [",
            *[f'    "{dep}",' for dep in deps],
            "]",
        ]
    )


def get_pyproject_toml_uv_sources(editable_dagster_root: Path) -> str:
    lib_lines = [
        f'{path.name} = {{ path = "{path}", editable = true }}'
        for path in _gather_dagster_packages(editable_dagster_root)
    ]
    return "\n".join(
        [
            "[tool.uv.sources]",
            *lib_lines,
        ]
    )


def _gather_dagster_packages(editable_dagster_root: Path) -> Sequence[Path]:
    return [
        p.parent
        for p in (
            *editable_dagster_root.glob("python_modules/dagster*/setup.py"),
            *editable_dagster_root.glob("python_modules/libraries/dagster*/setup.py"),
        )
    ]


def scaffold_code_location(
    path: Path,
    dg_context: DgContext,
    editable_dagster_root: Optional[str] = None,
    skip_venv: bool = False,
) -> None:
    click.echo(f"Creating a Dagster code location at {path}.")

    dependencies = get_pyproject_toml_dependencies(use_editable_dagster=bool(editable_dagster_root))
    dev_dependencies = get_pyproject_toml_dev_dependencies(
        use_editable_dagster=bool(editable_dagster_root)
    )
    uv_sources = (
        get_pyproject_toml_uv_sources(Path(editable_dagster_root)) if editable_dagster_root else ""
    )

    scaffold_subtree(
        path=path,
        name_placeholder="CODE_LOCATION_NAME_PLACEHOLDER",
        templates_path=os.path.join(
            os.path.dirname(__file__), "templates", "CODE_LOCATION_NAME_PLACEHOLDER"
        ),
        dependencies=dependencies,
        dev_dependencies=dev_dependencies,
        code_location_name=path.name,
        uv_sources=uv_sources,
    )

    # Build the venv
    cl_dg_context = dg_context.with_root_path(path)
    if cl_dg_context.use_dg_managed_environment and not skip_venv:
        cl_dg_context.ensure_uv_lock()
        RemoteComponentRegistry.from_dg_context(cl_dg_context)  # Populate the cache


# ########################
# ##### COMPONENT TYPE
# ########################


def scaffold_component_type(dg_context: DgContext, name: str) -> None:
    root_path = Path(dg_context.components_lib_path)
    click.echo(f"Creating a Dagster component type at {root_path}/{name}.py.")

    scaffold_subtree(
        path=root_path,
        name_placeholder="COMPONENT_TYPE_NAME_PLACEHOLDER",
        templates_path=os.path.join(os.path.dirname(__file__), "templates", "COMPONENT_TYPE"),
        project_name=name,
        component_type_class_name=camelcase(name),
        name=name,
    )

    with open(root_path / "__init__.py", "a") as f:
        f.write(f"from {dg_context.components_lib_package_name}.{name} import {camelcase(name)}\n")


# ########################
# ##### COMPONENT INSTANCE
# ########################


def scaffold_component_instance(
    path: Path,
    component_type: str,
    scaffold_params: Optional[Mapping[str, Any]],
    dg_context: "DgContext",
) -> None:
    click.echo(f"Creating a Dagster component instance folder at {path}.")
    os.makedirs(path, exist_ok=True)
    code_location_command = [
        "scaffold",
        "component",
        component_type,
        path,
        *(["--json-params", json.dumps(scaffold_params)] if scaffold_params else []),
    ]
    dg_context.external_components_command(code_location_command)



import json
import os
from pathlib import Path
from typing import Optional

import click

from dagster_dg.cli.global_options import dg_global_options
from dagster_dg.component import all_components_schema_from_dg_context
from dagster_dg.config import normalize_cli_config
from dagster_dg.context import DgContext
from dagster_dg.scaffold import scaffold_code_location
from dagster_dg.utils import DgClickCommand, DgClickGroup, exit_with_error
from dagster_dg.utils.vscode import (
    install_or_update_yaml_schema_extension,
    recommend_yaml_extension,
)


@click.group(name="code-location", cls=DgClickGroup)
def code_location_group():
    """Commands for operating code location directories."""


# ########################
# ##### SCAFFOLD
# ########################


@code_location_group.command(name="scaffold", cls=DgClickCommand)
@click.argument("name", type=str)
@click.option(
    "--use-editable-dagster",
    type=str,
    flag_value="TRUE",
    is_flag=False,
    default=None,
    help=(
        "Install Dagster package dependencies from a local Dagster clone. Accepts a path to local Dagster clone root or"
        " may be set as a flag (no value is passed). If set as a flag,"
        " the location of the local Dagster clone will be read from the `DAGSTER_GIT_REPO_DIR` environment variable."
    ),
)
@click.option(
    "--skip-venv",
    is_flag=True,
    default=False,
    help="Do not create a virtual environment for the code location.",
)
@dg_global_options
@click.pass_context
def code_location_scaffold_command(
    context: click.Context,
    name: str,
    use_editable_dagster: Optional[str],
    skip_venv: bool,
    **global_options: object,
) -> None:
    """Scaffold a Dagster code location file structure and a uv-managed virtual environment scoped
    to the code location.

    This command can be run inside or outside of a deployment directory. If run inside a deployment,
    the code location will be created within the deployment directory's code location directory.

    The code location file structure defines a Python package with some pre-existing internal
    structure:

    \b
    ├── <name>
    │   ├── __init__.py
    │   ├── components
    │   ├── definitions.py
    │   └── lib
    │       └── __init__.py
    ├── <name>_tests
    │   └── __init__.py
    └── pyproject.toml

    The `<name>.components` directory holds components (which can be created with `dg scaffold
    component`).  The `<name>.lib` directory holds custom component types scoped to the code
    location (which can be created with `dg component-type scaffold`).
    """  # noqa: D301
    cli_config = normalize_cli_config(global_options, context)
    dg_context = DgContext.from_config_file_discovery_and_cli_config(Path.cwd(), cli_config)
    if dg_context.is_deployment:
        if dg_context.has_code_location(name):
            exit_with_error(f"A code location named {name} already exists.")
        code_location_path = dg_context.code_location_root_path / name
    else:
        code_location_path = Path.cwd() / name

    if use_editable_dagster == "TRUE":
        if not os.environ.get("DAGSTER_GIT_REPO_DIR"):
            exit_with_error(
                "The `--use-editable-dagster` flag requires the `DAGSTER_GIT_REPO_DIR` environment variable to be set."
            )
        editable_dagster_root = os.environ["DAGSTER_GIT_REPO_DIR"]
    elif use_editable_dagster:  # a string value was passed
        editable_dagster_root = use_editable_dagster
    else:
        editable_dagster_root = None

    scaffold_code_location(
        code_location_path, dg_context, editable_dagster_root, skip_venv=skip_venv
    )


# ########################
# ##### LIST
# ########################


@code_location_group.command(name="list", cls=DgClickCommand)
@dg_global_options
@click.pass_context
def code_location_list_command(context: click.Context, **global_options: object) -> None:
    """List code locations in the current deployment."""
    cli_config = normalize_cli_config(global_options, context)
    dg_context = DgContext.for_deployment_environment(Path.cwd(), cli_config)

    for code_location in dg_context.get_code_location_names():
        click.echo(code_location)


_DEFAULT_SCHEMA_FOLDER_NAME = ".vscode"


@code_location_group.command(name="configure-vscode", cls=DgClickCommand)
@dg_global_options
@click.pass_context
def configure_vscode(
    context: click.Context,
    **global_options: object,
) -> None:
    """Generates and installs a VS Code extension which provides JSON schemas for Components types specified by YamlComponentsLoader objects."""
    cli_config = normalize_cli_config(global_options, context)
    dg_context = DgContext.from_config_file_discovery_and_cli_config(Path.cwd(), cli_config)

    recommend_yaml_extension()

    schema_folder = dg_context.root_path / _DEFAULT_SCHEMA_FOLDER_NAME
    schema_folder.mkdir(exist_ok=True)

    schema_path = schema_folder / "schema.json"
    schema_path.write_text(json.dumps(all_components_schema_from_dg_context(dg_context), indent=2))

    install_or_update_yaml_schema_extension(dg_context.root_path, schema_path)
